@@ -32,12 +32,6 @@ def _CK_MECHANISM_TYPE_to_enum(mechanism):
         return mechanism
 
 
-def _chunks(data, size):
-    """Yield successive `size`-sized chunks from `data`."""
-    for i in range(0, len(data), size):
-        yield data[i:i + size]
-
-
 class Slot:
     """
     A PKCS#11 device slot.
@@ -299,7 +293,7 @@ class Session:
 
         For generating keys see :meth:`generate_key` or
         :meth:`generate_keypair`.
-        For importing keys see :meth:`import_key` or :meth:`import_keypair`.
+        For importing keys see :meth:`import_key`.
 
         Requires a read/write session, unless the object is not to be
         stored.
@@ -340,6 +334,37 @@ class Session:
         """
         raise NotImplementedError()
 
+    def generate_keypair(self, key_type, key_length,
+                         id=None, label=None,
+                         store=True, capabilities=None,
+                         mechanism=None, mechanism_param=b'',
+                         public_template=None, private_template=None):
+        """
+        Generate a asymmetric keypair (e.g. RSA).
+
+        Keys should set at least `id` or `label`.
+
+        An appropriate `mechanism` will be chosen for `key_type`
+        (see :attr:`DEFAULT_GENERATE_MECHANISMS`) or this can be overridden.
+        Similarly the `capabilities` (see :attr:`DEFAULT_KEY_CAPABILITIES`).
+
+        The `template` will extend the default template used to make the
+        key.
+
+        :param KeyType key_type: Key type (e.g. KeyType.AES)
+        :param int key_length: Key length in bits (e.g. 256).
+        :param bytes id: Key identifier.
+        :param str label: Key label.
+        :param store: Store key on token (requires R/W session).
+        :param MechanismFlag capabilities: Key capabilities (or default).
+        :param Mechanism mechanism: Generation mechanism (or default).
+        :param bytes mechanism_param: Optional vector to the mechanism.
+        :param dict(Attribute,*) template: Additional attributes.
+
+        :rtype: (PublicKey, PrivateKey)
+        """
+        raise NotImplementedError()
+
     def seed_random(self, seed):
         """
         Mix additional seed material into the RNG (if supported).
@@ -457,14 +482,29 @@ class SecretKey(Key):
         return self[Attribute.VALUE_LEN] * 8
 
 
+class PublicKey(Key):
+    """
+    A PKCS#11 :attr:`pkcs11.constants.ObjectClass.PUBLIC_KEY` object
+    (asymmetric public key).
+    """
+
+    object_class = ObjectClass.PUBLIC_KEY
+
+
+class PrivateKey(Key):
+    """
+    A PKCS#11 :attr:`pkcs11.constants.ObjectClass.PRIVATE_KEY` object
+    (asymmetric private key).
+    """
+
+    object_class = ObjectClass.PRIVATE_KEY
+
+
 class EncryptMixin(Object):
     """
     This :class:`Object` supports the encrypt capability.
     """
 
-    def _encrypt(self, data, mechanism=None, mechanism_param=b''):
-        raise NotImplementedError()
-
     def encrypt(self, data, buffer_size=8192, **kwargs):
         """
         Encrypt some `data`.
@@ -475,7 +515,9 @@ class EncryptMixin(Object):
         (be aware, more chunks will be output than input).
 
         If you do not specify `mechanism` then the default from
-        :attr:`DEFAULT_ENCRYPT_MECHANISMS` will be used.
+        :attr:`DEFAULT_ENCRYPT_MECHANISMS` will be used. If an iterable
+        is passed and the mechanism chosen does not support handling data
+        in chunks, an exception will be raised.
 
         Some mechanisms (including the default CBC mechanisms) require an
         initialisation vector (of key length) to set the initial state of
@@ -483,12 +525,11 @@ class EncryptMixin(Object):
         vector should contain quality random. This method will not return
         the value of the initialisation parameter.
 
+        When passing an iterable for data
         `buffer_size` must be sufficient to store the working buffer. An
         integer number of blocks and greater than or equal to the largest
         input chunk is recommended.
 
-        Large single blocks of `data` will be chunked into the `buffer_size`.
-
         The returned generator obtains a lock on the :class:`Session`
         to prevent other threads from starting a simultaneous operation.
         The lock is released when you consume/destroy the generator.
@@ -500,13 +541,28 @@ class EncryptMixin(Object):
             by deleting the generator. You must consume the generator to
             complete the operation.
 
+        An example of streaming a file is as follows:
+
+        ::
+
+            def encrypt_file(file_in, file_out):
+
+                with \\
+                        open(file_in, 'rb') as input_, \\
+                        open(file_out, 'wb') as output:
+
+                    chunks = iter(lambda: input_.read(8192), '')
+
+                    for chunk in key.encrypt(chunks, mechanism_param=iv):
+                        output.write(chunk)
+
         :param data: data to encrypt
         :type data: str, bytes or iter(bytes)
         :param Mechanism mechanism: optional encryption mechanism
             (or None for default)
         :param bytes mechanism_param: optional mechanism parameter
             (e.g. initialisation vector).
-        :param int buffer_size: size of the working buffer.
+        :param int buffer_size: size of the working buffer (for generators)
 
         :rtype: bytes or iter(bytes)
         """
@@ -515,14 +571,12 @@ class EncryptMixin(Object):
         if isinstance(data, str):
             data = data.encode('utf-8')
 
-        # If we're not an iterable, recurse into ourselves with an iterable
-        # version and join the result at the end.
         if isinstance(data, bytes):
-            return b''.join(self._encrypt(_chunks(data, buffer_size),
-                                          buffer_size=buffer_size, **kwargs))
+            return self._encrypt(data, **kwargs)
 
         else:
-            return self._encrypt(data, buffer_size=buffer_size, **kwargs)
+            return self._encrypt_generator(data,
+                                           buffer_size=buffer_size, **kwargs)
 
 
 class DecryptMixin(Object):
@@ -530,9 +584,6 @@ class DecryptMixin(Object):
     This :class:`Object` supports the decrypt capability.
     """
 
-    def _decrypt(self, data, mechanism=None, mechanism_param=b''):
-        raise NotImplementedError()
-
     def decrypt(self, data, buffer_size=8192, **kwargs):
         """
         Decrypt some `data`.
@@ -545,27 +596,87 @@ class DecryptMixin(Object):
             (or None for default).
         :param bytes mechanism_param: optional mechanism parameter
             (e.g. initialisation vector).
-        :param int buffer_size: size of the working buffer.
+        :param int buffer_size: size of the working buffer (for generators).
 
         :rtype: bytes or iter(bytes)
         """
 
-        # If we're not an iterable, recurse into ourselves with an iterable
+        # If we're not an iterable, call into our generator with an iterable
         # version and join the result at the end.
         if isinstance(data, bytes):
-            return b''.join(self._decrypt(_chunks(data, buffer_size),
-                                          buffer_size=buffer_size, **kwargs))
+            return self._decrypt(data, **kwargs)
 
         else:
-            return self._decrypt(data, buffer_size=buffer_size, **kwargs)
+            return self._decrypt_generator(data,
+                                           buffer_size=buffer_size, **kwargs)
 
 
 class SignMixin(Object):
-    pass
+    """
+    This :class:`Object` supports the sign capability.
+    """
+
+    def sign(self, data, **kwargs):
+        """
+        Sign some `data`.
+
+        See :meth:`EncryptMixin.encrypt` for more information.
+
+        :param data: data to sign
+        :type data: str, bytes or iter(bytes)
+        :param Mechanism mechanism: optional signing mechanism
+        :param bytes mechanism_param: optional mechanism parameter
+
+        :rtype: bytes
+        """
+
+        # If data is a string, encode it now as UTF-8.
+        if isinstance(data, str):
+            data = data.encode('utf-8')
+
+        if isinstance(data, bytes):
+            return self._sign(data, **kwargs)
+
+        else:
+            return self._sign_generator(data, **kwargs)
 
 
 class VerifyMixin(Object):
-    pass
+    """
+    This :class:`Object` supports the verify capability.
+    """
+
+    def verify(self, data, signature, **kwargs):
+        """
+        Verify some `data`.
+
+        See :meth:`EncryptMixin.encrypt` for more information.
+
+        Returns True if `signature` is valid for `data`.
+
+        :param data: data to sign
+        :type data: str, bytes or iter(bytes)
+        :param bytes signature: signature
+        :param Mechanism mechanism: optional signing mechanism
+        :param bytes mechanism_param: optional mechanism parameter
+
+        :rtype: bool
+        """
+
+        # If data is a string, encode it now as UTF-8.
+        if isinstance(data, str):
+            data = data.encode('utf-8')
+
+        try:
+            if isinstance(data, bytes):
+                self._verify(data, signature, **kwargs)
+            else:
+                self._verify_generator(data, signature, **kwargs)
+
+            return True
+
+        except (SignatureInvalid, SignatureLenRange):
+            return False
 
 
 class WrapMixin(Object):
